#   coding=utf-8
#  #
#   Author: Ernesto Arredondo Martinez (ernestone@gmail.com)
#   File: oracle_pandas.py
#   Created: 05/04/2020, 17:41
#   Last modified: 10/11/2019, 11:24
#   Copyright (c) 2020

import geopandas as gpd
import pandas as pd
from pandas.api.types import CategoricalDtype
from functools import wraps

try:
    from apb_cx_oracle_spatial.gestor_oracle import sql_tab
    _HAS_ORACLE_DEPS = True
except ImportError:
    _HAS_ORACLE_DEPS = False

from . import optimize_df


def requires_oracle_deps(func):
    """Decorator to check if Oracle dependencies are installed"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not _HAS_ORACLE_DEPS:
            raise ImportError(
                "Oracle functionality requires additional dependencies. "
                "Install with 'pip install apb_pandas_utils[oracle]'"
            )
        return func(*args, **kwargs)
    return wrapper


@requires_oracle_deps
def df_for_sqlgen(a_ora_generator, columns_index=None, columns=None,
                  optimize=False,
                  **params_df):
    """
    A partir de generator de gestor_oracle devuelve pandas.Dataframe

    Args:
        a_ora_generator:
        columns_index:
        columns: lista con nombre de columnas a mostrar
        optimize: indica si se intentará optimizar el espacio ocupado por el Dataframe
        params_df: parametros creación Dataframe

    Returns:
        pandas.Dataframe
    """
    df = pd.DataFrame(a_ora_generator, **params_df)
    if not df.empty and columns_index:
        df.set_index([c.upper() for c in columns_index], inplace=True)

    if columns:
        df = df[[c.upper() for c in columns if c in df.columns]]

    if optimize:
        df = optimize_df(df)

    return df


@requires_oracle_deps
def gdf_for_sqlgen(a_ora_generator, column_geom, crs=None, **params_df_for_sqlgen):
    """
    A partir de generator de gestor_oracle devuelve geopandas.Dataframe

    Args:
        a_ora_generator:
        column_geom:
        crs (str=None): name EPSG crs
        **params_df_for_sqlgen: parametros opcionales funcion df_for_sqlgen()

    Returns:
        geopandas.Dataframe
    """
    column_geom = column_geom.upper()
    cols = params_df_for_sqlgen.pop('columns', [])
    if cols:
        params_df_for_sqlgen['columns'] = cols_sql(cols, [column_geom])

    gdf = gpd.GeoDataFrame(df_for_sqlgen(a_ora_generator,
                                         **params_df_for_sqlgen),
                           geometry=column_geom,
                           crs=crs)

    return gdf


@requires_oracle_deps
def cols_sql(cols, *l_cols_add):
    """
    Retorna lista de columnas

    Args:
        cols (list):
        *l_cols_add: listas de columnas a añadir

    Returns:
        list
    """
    if cols:
        cols = [c.upper() for c in cols]
        for added_cols in l_cols_add:
            cols += list(map(str.upper, added_cols))

        return list(set(cols))


@requires_oracle_deps
def df_table(gest_ora, nom_tab, filter_sql=None, *args_filter_sql, **params_df_for_sqlgen):
    """
    Devuelve pandas.DataFrame para una tablas de oracle

    Args:
        gest_ora:
        nom_tab:
        filter_sql:
        *args_filter_sql:
        **params_df_for_sqlgen: parametros opcionales funcion df_for_sqlgen()

    Returns:
        pandas.Dataframe
    """
    dd_tab = gest_ora.get_dd_table(nom_tab)
    cols = params_df_for_sqlgen.get("columns", []).copy()
    cols_idx = params_df_for_sqlgen.pop('columns_index', dd_tab.pk())

    df_tab = df_for_sqlgen(
        gest_ora.generator_rows_sql(sql_tab(nom_tab,
                                            filter_sql,
                                            cols_sql(cols,
                                                     cols_idx)),
                                    *args_filter_sql,
                                    geom_format="as_geojson"),
        columns_index=cols_idx,
        **params_df_for_sqlgen)

    if not cols:
        df_tab = df_tab[[c for c in dd_tab.cols if c in df_tab.columns]]

    return df_tab


@requires_oracle_deps
def gdf_table(gest_ora, nom_tab, column_geom, other_cols_geom=False, null_geoms=False,
              filter_sql=None, *args_filter_sql, **params_gdf_for_sqlgen):
    """
    Devuelve geopandas.GeoDataframe para una tablas de oracle

    Args:
        gest_ora:
        nom_tab:
        column_geom:
        other_cols_geom:
        null_geoms:
        filter_sql:
        *args_filter_sql:
        **params_gdf_for_sqlgen:

    Returns:
        geopandas.GeoDataframe
    """
    dd = gest_ora.get_dd_table(nom_tab)
    column_geom = column_geom.upper()
    cols = params_gdf_for_sqlgen.pop('columns', [])
    if not other_cols_geom:
        if not cols:
            cols = [c for c in dd.alfas(include_pk=False)]
        else:
            other_geoms = [g for g in dd.geoms().keys() if g != column_geom.upper()]
            cols = [c for c in cols
                    if c.upper() not in other_geoms]

    cols_idx = params_gdf_for_sqlgen.pop('columns_index', dd.pk())

    if not null_geoms:
        not_null_geoms_sql = "{} is not null".format(column_geom)
        if filter_sql:
            filter_sql = "{} and ({})".format(not_null_geoms_sql, filter_sql)
        else:
            filter_sql = not_null_geoms_sql

    crs_epsg = None
    if (tip_geom := gest_ora.get_tip_camp_geom(nom_tab, column_geom)) and \
            (epsg_code := gest_ora.get_epsg_for_srid(tip_geom.SRID)):
        crs_epsg = f'EPSG:{epsg_code}'

    gdf_tab = gdf_for_sqlgen(gest_ora.generator_rows_sql(sql_tab(nom_tab,
                                                                 filter_sql,
                                                                 cols_sql(cols,
                                                                          cols_idx,
                                                                          [column_geom])),
                                                         *args_filter_sql,
                                                         geom_format="as_shapely"),
                             column_geom=column_geom,
                             crs=crs_epsg,
                             columns_index=cols_idx,
                             columns=cols, **params_gdf_for_sqlgen)

    if not cols:
        gdf_tab = gdf_tab[[c for c in dd.cols if c in gdf_tab.columns]]

    return gdf_tab


@requires_oracle_deps
def dtype_categorical_from_tab_column(gest_ora, nom_tab, nom_col):
    """
    Retorna pandas.dtype de tipo categorico para la columna de una tabla

    Args:
        gest_ora:
        nom_tab:
        nom_col:

    Returns:
        pandas.CategoricalDtype
    """
    return CategoricalDtype(categories=gest_ora.iter_distinct_vals_camp_tab(nom_tab, nom_col))



#  coding=utf-8
#
#  Author: Ernesto Arredondo Martinez (ernestone@gmail.com)
#  Created: 
#  Copyright (c)
"""
.. include:: ../README.md
"""
from __future__ import annotations

import re
from collections.abc import Iterable
from datetime import datetime, time, date
from typing import Union

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from pandas import DataFrame, Timestamp, NaT, CategoricalDtype

EXCLUDED_TYPES_TO_CATEGORIZE = ['datetime', 'category', 'geometry']
DEFAULT_MAX_UNIQUE_VALS_COL_CATEGORY = 0.5
MAX_DATETIME = datetime(2250, 12, 31, 23, 59, 59)


def optimize_df(df: DataFrame | GeoDataFrame, max_perc_unique_vals: float = DEFAULT_MAX_UNIQUE_VALS_COL_CATEGORY,
                floats_as_categ: bool = False) -> DataFrame | GeoDataFrame:
    """
    Retorna el pd.Dataframe optimizado segun columnas que encuentre

    Args:
        df (Dataframe | GeoDataFrame): Dataframe a optimizar
        max_perc_unique_vals (float=DEFAULT_MAX_UNIQUE_VALS_COL_CATEGORY): Màxim percentatge de valors únics respecte total files per a convertir en categoria, expressat entre 0 i 1 (Default 0.5 -> 50%)
        floats_as_categ (bool=False): Si True, els floats es converteixen a categoria

    Returns:
        opt_df (Dataframe | GeoDataFrame): Dataframe optimizado
    """
    opt_df = df.copy()
    df_ints = opt_df.select_dtypes(include=['int64'])
    opt_df[df_ints.columns] = df_ints.apply(pd.to_numeric, downcast='signed')
    df_floats = opt_df.select_dtypes(include='float')
    opt_df[df_floats.columns] = df_floats.apply(pd.to_numeric, downcast='float')

    excl_types_cat = EXCLUDED_TYPES_TO_CATEGORIZE
    if not floats_as_categ:
        excl_types_cat.append('float')

    for col in opt_df.select_dtypes(exclude=excl_types_cat).columns:
        try:
            unic_vals = opt_df[col].unique()
        except (pd.errors.DataError, TypeError):
            continue

        num_unique_values = len(unic_vals)
        num_total_values = len(opt_df[col]) - len(opt_df.loc[opt_df[col].isnull()])
        if num_total_values > 0 and (num_unique_values / num_total_values) < max_perc_unique_vals:
            try:
                opt_df[col] = opt_df[col].astype(CategoricalDtype(ordered=True))
            except (NotImplementedError, TypeError):
                continue

    return opt_df


def df_filtered_by_prop(df: DataFrame | GeoDataFrame, filter_prop: dict[str, object]) -> DataFrame | GeoDataFrame | None:
    """
    Filtra el dataframe amb el diccionari passat, on la clau fa referència a la columna i el valor o llistat de valors
    separats per comes son els que s’apliquen al filtre. Si la clau/columna no existeix es desestima. Si la clau/columna
    comença per alguns d’aquest signes “=, !, -, >, <” s’aplica la corresponent operació de filtre.
    En el cas de “!” i “–“ s’aplica la mateixa operació de negat o que no contingui el valor o valors passats.
    Els filtres “<“ i “>” no apliquen a camps text i es desestimen. Es poden passar la mateixa columna amb operadors
    i valors distints per aplicar filtres diferents

    Args:
        df (DataFrame | GeoDataFrame): DataFrame a filtrar
        filter_prop (dict[str, object]): Propietats de filtrat

    Returns:
        DataFrame | GeoDataFrame: DataFrame filtrat
    """
    if df is None or not filter_prop:
        return df

    idx_names = [idx_col for idx_col in df.index.names if idx_col]
    if idx_names:
        df = df.reset_index()

    def _df_individual_filter(_df_ind: DataFrame, type_col_ind, column: str, value, col_operator: str = '='):
        type_column = type_col_ind.categories.dtype if (type_col_name := type_col_ind.name) == 'category' else type_col_ind

        if type_col_name == 'object':
            if col_operator == '=':
                _df_ind = _df_ind[_df_ind[column].str.contains(str(value), case=False, na=False)]
            elif col_operator == '-' or col_operator == '!':
                _df_ind = _df_ind[~_df_ind[column].str.contains(str(value), case=False, na=False)]
        else:
            value = type_column.type(value)
            if col_operator == '=':
                _df_ind = _df_ind.loc[_df_ind[column] == value]
            elif col_operator == '-' or col_operator == '!':
                _df_ind = _df_ind.loc[_df_ind[column] != value]
            elif col_operator == '>':
                _df_ind = _df_ind.loc[_df_ind[column] > value]
            elif col_operator == '<':
                _df_ind = _df_ind.loc[_df_ind[column] < value]

        return _df_ind

    col_names = df.columns.values.tolist()
    for k, v in filter_prop.items():
        k_operator = "="
        if k.startswith(('-', '=', '<', '>', '!')):
            k_operator = k[0:1]
            k = k[1:]
        if k.upper() in (col_names + idx_names):
            k = k.upper()
        elif k.lower() in (col_names + idx_names):
            k = k.lower()

        if k in col_names and v is not None:
            type_col = df.dtypes.get(k)
            if isinstance(v, list):
                # es fa amb un bucle i no amb isin perque no val per floats
                df_list = None
                for ind_val in v:
                    df_temp = _df_individual_filter(df, type_col, k, ind_val, k_operator)
                    if df_list is None:
                        df_list = df_temp
                    elif k_operator == '=':
                        df_list = pd.concat([df_list, df_temp])
                    if k_operator != '=':
                        # per als operadors que exclouen s'ha de treballar sobre el df filtrat resultant
                        df = df_list = df_temp
                df = df_list
            else:
                df = _df_individual_filter(df, type_col, k, v, k_operator)

    if idx_names:
        df.set_index(idx_names, inplace=True)

    return df


def rename_and_drop_columns(df: Union[DataFrame, GeoDataFrame], map_old_new_col_names: dict[str, str],
                            drop_col: bool = True, strict: bool = False, reordered: bool = False) -> Union[
    DataFrame, GeoDataFrame]:
    """
    Function to rename and remove columns from a dataframe. If the drop_col parameter is True,
    the columns that are not in the map will be removed. If the strict parameter is True,
    the names that do not exist in the map as a column will be skipped.
    Args:
        df: to remove and rename
        map_old_new_col_names: the key is the actual name and the value is the new name
        drop_col: True to remove columns that are not included in the map
        strict: False to skip names that are not included in the map
        reordered: True to reorder columns of dataframe

    Returns: modified DataFrame

    """
    if df is not None and map_old_new_col_names:
        col_names = df.columns.values.tolist()
        col_names_to_drop = col_names.copy()
        final_map = {}
        for k, v in map_old_new_col_names.items():
            if k in col_names:
                final_map[k] = v
                col_names_to_drop.remove(k)
        if drop_col:
            df = df.drop(col_names_to_drop, axis=1)
        if strict:
            final_map = map_old_new_col_names
        df = df.rename(columns=final_map)
        if reordered:
            new_cols = list(map_old_new_col_names.values())
            act_cols = df.columns.tolist()
            reord_cols = [value for value in new_cols if value in act_cols]
            df = df[reord_cols]
        return df


def set_null_and_default_values(df: DataFrame | GeoDataFrame) -> DataFrame | GeoDataFrame:
    """
    Function to replace NaN values with None in a DataFrame
    Args:
        df (DataFrame | GeoDataFrame): DataFrame to replace NaN values with None

    Returns:
        DataFrame | GeoDataFrame: DataFrame with NaN values replaced with None
    """
    df = df.replace({np.nan: None})
    return df


def replace_values_with_null(df: Union[DataFrame | GeoDataFrame], dict_col_values: dict) -> Union[
    DataFrame | GeoDataFrame]:
    """
    Function to replace values with None in a DataFrame
    Args:
        df (DataFrame | GeoDataFrame): DataFrame to replace values with None
        dict_col_values (dict): Dictionary with the column name and the value to replace with None

    Returns:
        DataFrame | GeoDataFrame: DataFrame with values replaced with None
    """
    if df is not None and not df.empty and dict_col_values:
        for name_col, value in dict_col_values.items():
            df[name_col] = df[name_col].replace(value, None)
    return df


def convert_to_datetime_col_df(df: DataFrame, cols: list[str],
                               set_end_day: bool = False, set_nat: bool = False) -> DataFrame:
    """
    Force convert date columns to datetime format.
    If init_date is True, the time is set to 00:00:00 if not to 23:59:59

    Args:
        df (DataFrame): DataFrame
        cols (list[str]): Columns to convert
        set_end_day (bool=False): If False the time is set to 00:00:00 if not to 23:59:59
        set_nat (bool=False): If True set NaT to MAX_DATETIME

    Returns:
        DataFrame: DataFrame with datetime columns
    """
    if not set_end_day:
        delta_time = time.min
    else:
        delta_time = time(23, 59, 59)

    def _convert_date(value):
        if type(value) is date:
            value = datetime.combine(value, time.min)

        if set_nat and (value is NaT or value is None):
            return MAX_DATETIME
        elif value is NaT:
            return value
        elif ((isinstance(value, Timestamp) or isinstance(value, datetime))
              and set_end_day and value.time() == time.min):
            return datetime.combine(value, delta_time)
        else:
            return value

    for col in cols:
        df[col] = df[col].apply(_convert_date)

    return df


def df_memory_usage(df: DataFrame | GeoDataFrame) -> float:
    """
    Return the memory usage of a DataFrame in MB

    Args:
        df (DataFrame | GeoDataFrame): DataFrame

    Returns:
        float: Memory usage in MB
    """
    return df.memory_usage(deep=True).sum() / 1024 ** 2


def extract_operator(input_string):
    """
    Extract sql operator from input string

    Args:
        input_string:

    Returns:

    """
    special_opers = ('-', '!') # Special operators for negation
    sql_opers = ('=', '!=', '<>', '>', '<', '>=', '<=') # SQL operators

    match = re.match(r'^(\W+)', input_string)
    if match:
        symbols = match.group(1)

        if symbols not in special_opers and symbols not in sql_opers:
            raise ValueError(f"Operator '{symbols}' not supported")

        return symbols

    return None


def sql_from_filter_by_props(**filter_by_props: dict) -> str:
    """
    Get SQL from filter by properties

    Args:
        **filter_by_props: The filter by properties

    Returns:
        sql (str): The SQL from filter by properties
    """
    sql_parts = []
    for k_fld_oper, value in filter_by_props.items():
        if k_operator := extract_operator(k_fld_oper):
            k_fld = k_fld_oper.replace(k_operator, '')
        else:
            k_operator = "="
            k_fld = k_fld_oper

        if isinstance(value, str):
            value = f"'{value}'"
        elif isinstance(value, Iterable):
            value = ', '.join([f"'{v}'" if isinstance(v, str) else str(v) for v in value])
            value = f"({value})"
            if k_operator in ('=', '!=', '-'):
                k_operator = "IN" if k_operator == "=" else "NOT IN"
            else:
                raise ValueError(f"Operator '{k_operator}' not supported for iterable values")

        sql_parts.append(f"{k_fld} {k_operator} {value}")

    sql_filter = ' AND '.join(sql_parts)

    return sql_filter



import datetime
import unittest
import os

import geopandas as gpd
import pandas as pd
import requests
from geopandas import GeoDataFrame

from apb_extra_utils.misc import unzip
from apb_extra_utils.utils_logging import get_base_logger
from apb_pandas_utils import df_memory_usage
from apb_pandas_utils.geopandas_utils import gdf_to_geojson, gdf_from_df, gdf_to_df, df_to_crs, gdf_from_url

RESOURCES_DATA_DIR = os.path.join(
    os.path.dirname(
        os.path.dirname(os.path.dirname(__file__))),
    'resources', 'data')


class GeopandasUtilsTestCase(unittest.TestCase):
    unzip(os.path.join(RESOURCES_DATA_DIR, 'edificacio.zip'))
    csv_path = os.path.join(RESOURCES_DATA_DIR, 'edificacio', 'edificacio.csv')
    geojson_path = os.path.join(RESOURCES_DATA_DIR, 'edificacio-perimetre_base.geo.json')
    logger = get_base_logger()

    def setUp(self):
        self.gdf_json = gpd.read_file(self.geojson_path, engine='pyogrio')
        self.df_csv = pd.read_csv(self.csv_path)

    def test_gdf_to_geojson(self):
        self.logger.info('Converting GeoDataFrame to geojson')
        dict_geo = gdf_to_geojson(
            self.gdf_json,
            'Test geojson',
            with_crs=True,
            show_bbox=True,
            drop_id=False)
        self.logger.info(f'Geojson: {dict_geo}')

    def test_gdf_from_df(self):
        self.logger.info('Converting DataFrame to GeoDataFrame')
        gdf = gdf_from_df(self.df_csv, geom_col='PERIMETRE_BASE', crs='EPSG:4326',
                          cols_geom=['PERIMETRE_SUPERIOR', 'PUNT_BASE', 'DENOMINACIO'])
        gdf_epsg25831 = df_to_crs(gdf, 'EPSG:25831')
        self.logger.info(f'GeoDataFrame: {gdf_epsg25831.shape} | Memory: {df_memory_usage(gdf_epsg25831):.2f} MB')

        self.logger.info('Converting DataFrame with Index to GeoDataFrame')
        df_csv_idx = self.df_csv.copy().set_index('APB_ID')
        gdf = gdf_from_df(df_csv_idx, geom_col='PERIMETRE_BASE', crs='EPSG:4326',
                          cols_geom=['PERIMETRE_SUPERIOR', 'PUNT_BASE', 'DENOMINACIO'])
        self.logger.info(
            f'GeoDataFrame: {gdf.shape} with index {gdf.index.name} | Memory: {df_memory_usage(gdf_epsg25831):.2f} MB')

    def test_gdf_to_dataframe(self):
        self.logger.info('Converting GeoDataFrame to DataFrame as WKT')
        gdf_csv = gdf_from_df(self.df_csv, geom_col='PUNT_BASE', crs='EPSG:4326',
                              cols_geom=['PERIMETRE_SUPERIOR', 'PUNT_BASE', 'DENOMINACIO', 'PERIMETRE_BASE'])
        df = gdf_to_df(gdf_csv)
        self.logger.info(
            f'DF from CSV: {self.df_csv.shape} <> DF from GDF: {df.shape} | Memory: {df_memory_usage(df):.2f} MB')

        self.logger.info('Converting GeoDataFrame to DataFrame as WKB')
        df_wkb = gdf_to_df(gdf_csv, as_wkb=True)
        self.logger.info(f'DF from GDF with WKB: {df_wkb.shape} | Memory: {df_memory_usage(df_wkb):.2f} MB')

    def test_gdf_from_url_repo_gis(self):
        self.logger.info('Loading GeoDataFrame from URL Repo GIS')
        url_geojson = 'https://gisplanoldev.portdebarcelona.cat/repo_gis_pg/api/current/edificacio-perimetre_base'
        gdf_url = gdf_from_url(url_geojson, crs_api='EPSG:25831', add_goto_url=True)
        self.logger.info(f'GeoDataFrame from URL: {gdf_url.shape} | Memory: {df_memory_usage(gdf_url):.2f} MB')

    def test_gdf_from_url_escales(self):
        self.logger.info('Loading GeoDataFrame from URL Escales')
        resp_auth = requests.post(
            'https://gisplanol.portdebarcelona.cat/adm_escales/auth/acces-token',
            json=dict(
                username=os.getenv('DJANGO_API_USER'),
                key=os.getenv('DJANGO_API_KEY')
            )
        )
        url_geojson = 'https://gisplanol.portdebarcelona.cat/adm_escales/api/track-ship-position/'
        api_params = dict(start=datetime.datetime.now().strftime('%Y-%m-%d'), limit=100)
        headers = {'Authorization': f'Bearer {resp_auth.json().get("access")}'}
        gdf_url: GeoDataFrame = gdf_from_url(
            url_geojson,
            api_params=api_params,
            crs_api='EPSG:4326',
            headers=headers,
            add_goto_url=True)
        self.logger.info(f'GeoDataFrame from URL: {gdf_url.shape} | Memory: {df_memory_usage(gdf_url):.2f} MB')
        gdf_url.to_file(filename=os.path.join(RESOURCES_DATA_DIR, 'track-ship-position.geojson'), driver='GeoJSON')


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import pandas as pd

from apb_extra_utils.misc import unzip
from apb_extra_utils.utils_logging import get_base_logger
from apb_pandas_utils import df_memory_usage, optimize_df, df_filtered_by_prop

RESOURCES_DATA_DIR = os.path.join(
    os.path.dirname(
        os.path.dirname(os.path.dirname(__file__))),
    'resources', 'data')


class PandasUtilsTestCase(unittest.TestCase):
    unzip(os.path.join(RESOURCES_DATA_DIR, 'edificacio.zip'))
    csv_path = os.path.join(RESOURCES_DATA_DIR, 'edificacio', 'edificacio.csv')
    logger = get_base_logger()

    def setUp(self):
        self.df_csv = pd.read_csv(self.csv_path)
        self.df_csv.set_index('APB_ID', inplace=True)

    def test_optimize_df(self):
        self.logger.info('Optimize DataFrame')
        self.logger.info(f'DF before: {self.df_csv.shape} | Memory: {df_memory_usage(self.df_csv):.2f} MB')
        df_opt = optimize_df(self.df_csv)
        cols_categ = df_opt.select_dtypes(include=['category']).columns.tolist()
        self.logger.info(f'Columns: {cols_categ}')
        self.logger.info(f'DF after: {df_opt.shape} | Memory: {df_memory_usage(df_opt):.2f} MB')

    def test_filtered_by_prop(self):
        self.logger.info('Filter DataFrame by properties')
        df_csv = self.df_csv
        filter_props = {
            '>=NUMERO_PLANTES': 1,
            'SEQENTITAT': [250494, 250488, 250829],
            'TIPUS_EDIFICACIO': ['Edificació secundària o auxiliar', 'Edifici principal'],
        }
        df_filtered = df_filtered_by_prop(df_csv, filter_props)
        self.logger.info(f'DF filtered without categories: {df_filtered.shape} | '
                         f'Memory: {df_memory_usage(df_filtered):.2f} MB')

        df_cat = optimize_df(df_csv)
        df_filtered_cat = df_filtered_by_prop(df_cat, filter_props)
        self.logger.info(f'DF filtered with categories: {df_filtered_cat.shape} | '
                         f'Memory: {df_memory_usage(df_filtered_cat):.2f} MB')
        self.assertEqual(df_filtered_cat.shape[0], 3)

        # EAM - Test para comprobar que rango de valores (NUMERO_PLANTES) NOO filtra correctamente!!!
        filter_props['<=NUMERO_PLANTES'] = 2
        df_filtered_cat = df_filtered_by_prop(df_cat, filter_props)
        self.assertEqual(df_filtered_cat.shape[0], 3)


if __name__ == '__main__':
    unittest.main()



#   coding=utf-8
#  #
#   Author: Ernesto Arredondo Martinez (ernestone@gmail.com)
#   File: setup.py
#   Created: 05/04/2020, 00:21
#   Last modified: 05/04/2020, 00:21
#   Copyright (c) 2020
import importlib
import os

from setuptools import setup, find_packages


GIT_REPO = 'https://github.com/portdebarcelona/PLANOL-generic_python_packages'


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='apb_pandas_utils',
    version='1.2.0',
    packages=find_packages(),
    url=f'{GIT_REPO}/tree/master/apb_pandas_utils_pckg',
    author='Ernesto Arredondo Martínez',
    author_email='ernestone@gmail.com',
    maintainer='Port de Barcelona',
    maintainer_email='planolport@portdebarcelona.cat',
    description='Pandas and geopandas utils',
    long_description=readme(),
    # Ver posibles clasifiers aqui [https://pypi.org/classifiers/]
    classifiers=[
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
        'Operating System :: OS Independent'
    ],
    install_requires=[
        'geopandas>=1.0',
        'requests'
    ],
    extras_require={
        'oracle': ['apb_cx_oracle_spatial<1.2']
    },
    python_requires='>=3.6',
    package_data={
        # If any package contains *.txt, *.md or *.yml files, include them:
        "": ["*.txt", "*.md", "*.yml"]
    }
)



#  coding=utf-8
#
#  Author: Ernesto Arredondo Martinez (ernestone@gmail.com)
#  Created: 7/6/19 18:23
#  Last modified: 7/6/19 18:21
#  Copyright (c) 2019
import json
from typing import Optional

import requests
from geopandas import GeoDataFrame, GeoSeries
from pandas import DataFrame, Series
from shapely import wkt


def gdf_to_geojson(gdf: GeoDataFrame, name: Optional[str] = None, with_crs: bool = True, show_bbox: bool = True,
                   drop_id: bool = False, path_file: str = None) -> dict:
    """
    Convierte un GeoDataFrame a diccionario geojson

    Args:
        gdf (GeoDataFrame):
        name (str=None):
        with_crs (bool=True):
        show_bbox (bool=True):
        drop_id (bool=False):
        path_file (str=None): Si se indica se guarda el geojson en el path indicado

    Returns:
        dict_geojson (dict)
    """
    dict_geojson = gdf.to_geo_dict(show_bbox=show_bbox, drop_id=drop_id)
    if name:
        dict_geojson["name"] = name
    if with_crs and gdf.crs is not None:
        auth = gdf.crs.to_authority()
        dict_geojson["crs"] = {"type": "name", "properties": {"name": f"urn:ogc:def:crs:{auth[0]}::{auth[1]}"}}

    if path_file:
        geojson = json.dumps(dict_geojson, default=str, ensure_ascii=False)
        with open(path_file, 'w', encoding='utf-8') as f:
            f.write(geojson)

    return dict_geojson


def gdf_to_df(gdf: GeoDataFrame, as_wkb=False) -> DataFrame:
    """
    Convert a GeoDataFrame to DataFrame converting the geometry columns to a str column in WKT format (WKB if as_wkb=True)

    Args:
        gdf (GeoDataFrame):
        as_wkb (bool=False): If True, the geometry column is converted to WKB format

    Returns:
        DataFrame
    """
    f_conv = 'to_wkb' if as_wkb else 'to_wkt'

    # Convert all columns type geometry to WKT
    gdf_aux = gdf.copy()
    for col in df_geometry_columns(gdf_aux):
        gdf_aux[col] = getattr(gdf_aux[col], f_conv)()
    return DataFrame(gdf_aux)


def df_geometry_columns(df: GeoDataFrame | DataFrame) -> list:
    """
    Devuelve las columnas tipo geometría de un GeoDataFrame

    Args:
        df (GeoDataFrame | DataFrame):

    Returns:
        list
    """
    return df.select_dtypes(include=["geometry"]).columns.tolist()


def df_to_crs(df: GeoDataFrame | DataFrame, crs: str) -> GeoDataFrame | DataFrame:
    """
    Convierte todas las columnas tipo geometría de un GeoDataFrame o DataFrame al CRS indicado

    Args:
        df (GeoDataFrame | DataFrame):
        crs (str): name CRS (EPSG) coord .sys. destino de las geometrías (e.g. 'EPSG:25831')
                    [Can be anything accepted by pyproj.CRS.from_user_input()]

    Returns:
        GeoDataFrame | DataFrame
    """
    df_aux = df.copy()
    for geom in df_geometry_columns(df_aux):
        df_aux[geom] = df_aux[geom].to_crs(crs)

    df_aux = df_aux.to_crs(crs)

    return df_aux


def gdf_from_df(df: DataFrame, geom_col: str, crs: str, cols_geom: list[str] = None) -> GeoDataFrame:
    """
    Crea un GeoDataFrame a partir de un DataFrame

    Args:
        df (DataFrame):
        geom_col (str): Columna geometría con el que se creará el GeoDataFrame
        crs (str): CRS (EPSG) coord .sys. origen de las geometrías (e.g. 'EPSG:25831')
                    [Can be anything accepted by pyproj.CRS.from_user_input()]
        cols_geom (list=None): Columnas con geometrías

    Returns:
        GeoDataFrame
    """
    if cols_geom is None:
        cols_geom = []

    cols_geom = set(cols_geom)
    cols_geom.add(geom_col)

    df_aux = df.copy()
    idx_prev = df_aux.index
    # We only deal with index when has names setted referred to possible columns
    set_idx = None not in idx_prev.names
    if set_idx:
        df_aux.reset_index(inplace=True)

    def convert_to_wkt(val_col):
        return wkt.loads(val_col) if isinstance(val_col, str) else None

    gdf = GeoDataFrame(df_aux)
    for col in (col for col in gdf.columns if col in cols_geom):
        ds_col = gdf[col]
        if isinstance(ds_col, GeoSeries):
            continue

        if (dtype := ds_col.dtype.name) == 'object':
            gdf[col] = gdf[col].apply(convert_to_wkt)

        gdf.set_geometry(col, inplace=True, crs=crs)

    if set_idx:
        gdf = gdf.set_index(idx_prev.names, drop=True)

    gdf.set_geometry(geom_col, crs=crs, inplace=True)

    return gdf


def gdf_from_url(url_rest_api, api_params=None, crs_api=None, headers=None, crs_gdf=None, add_goto_url=False):
    """
    Fetch paginated GeoJSON from a REST API and return a GeoPandas GeoDataFrame.

    Assumes the API returns a GeoJSON FeatureCollection with 'features' and optionally 'next' for pagination.
    If 'next' is present, it should be the full URL for the next page.

    Args:
        url_rest_api (str): The base URL of the API endpoint.
        api_params (dict, optional): Query parameters for the initial request.
        crs_api (str, optional): CRS (EPSG) coord .sys. origen de las geometrías (e.g. 'EPSG:25831')
                    [Can be anything accepted by pyproj.CRS.from_user_input()]
        headers (dict, optional): HTTP headers for the request.
        crs_gdf (str, optional): CRS (EPSG) coord .sys. destino de las geometrías (e.g. 'EPSG:25831')
                    [Can be anything accepted by pyproj.CRS.from_user_input()]
        add_goto_url (bool, optional): If True, adds a 'goto_url' to the GeoDataFrame as new column.

    Returns:
        gpd.GeoDataFrame | None: A GeoDataFrame containing all features from all pages.

    Raises:
        requests.HTTPError: If any request fails.
    """
    gdf = None
    all_features = []
    url = url_rest_api
    params = api_params or {}
    first_request = True

    while url:
        if first_request:
            response = requests.get(url, params=params, headers=headers)
            first_request = False
        else:
            response = requests.get(url, headers=headers)

        response.raise_for_status()
        data = response.json()

        # Assuming GeoJSON FeatureCollection. Test results or data directly
        all_features.extend(data.get('results', data).get('features', []))

        # Check for next page
        url = data.get('next')

    # Create GeoDataFrame from all features
    if all_features:
        gdf = GeoDataFrame.from_features(all_features, crs=crs_api)

    if add_goto_url:
        centroids = gdf.geometry.centroid.to_crs('EPSG:4326')
        mask = centroids.notna()
        gdf['goto_url'] = Series([None] * len(gdf), index=gdf.index)
        gdf.loc[mask, 'goto_url'] = \
            ("https://www.google.com/maps?q=" +
             centroids.loc[mask].y.astype(str) + "," +
             centroids.loc[mask].x.astype(str))

    if crs_gdf:
        gdf = gdf.to_crs(crs_gdf)

    return gdf
